/*********************************************
 * Author: Clément Notin for "Erasme - Département du Rhône" (France) in July 2008
 * Email : clement@notin.org
 * Website : reseau.erasme.org
 * Based on : Guido Socher's code example found on http://www.tuxgraphics.com/electronics/200611/embedded-webserver.shtml (end of page)
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Main program : initiate network connection (UDP, ARP and ICMP (ping only), NO TCP) and serial connection (with "binem carrier board")
 * This prog receives commands (3 types) via UDP and sends it to "binem carrier board" over serial
 *
 * Chip type  : ATMEGA168 with ENC28J60
 *********************************************/

#include <avr/interrupt.h> //sei()
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h" //delay_ms()
#include "avr_compat.h"
#include "net.h"
#include "uart.h"

#define F_CPU 12500000UL //12.5 MHz
#define BAUDRATE 38400 //higher baudrates have been tested : magjack's clock is not precise enough for a reliable transfer

#define ROW_LEN 240 //how long is a row in bytes (60 : HVGA format)

// please modify the following two lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x24};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
static uint8_t myip[4] = {192,168,0,144};

// listen port for udp
#define MYUDPPORT 1200

#define BUFFER_SIZE 450
static uint8_t buf[BUFFER_SIZE+1];

#define TRUE 0
#define FALSE 1


int main(void)
{
    uint16_t plen;
    uint8_t i=0; //dummy counter for looping on all bytes of a row
    uint8_t payloadlen=0;
    uint8_t ready=TRUE;
    uint8_t packetCount=0;
    // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
    // full external speed)
    // set the clock prescaler. First write CLKPCE to enable setting of clock the
    // next four instructions.
    CLKPR=(1<<CLKPCE); // change enable
    CLKPR=0; // "no pre-scaler"
    delay_ms(1);

    /* enable PD2/INT0, as input */
    DDRD&= ~(1<<DDD2);

    /*initialize enc28j60*/
    enc28j60Init(mymac);
    enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
    delay_ms(10);

    // LED
    /* enable PB1, LED as output */
    DDRB|= (1<<DDB1);


    // the transistor on PD7 : UNUSED but kept for example...
    DDRD|= (1<<DDD7);
    PORTD &= ~(1<<PD7);// transistor off

    /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
    // LEDB=yellow LEDA=green
    //
    // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
    // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
    enc28j60PhyWrite(PHLCON,0x476);
    delay_ms(20);

    /* set output to GND, red LED on */
    PORTB &= ~(1<<PB1);

    //init the ethernet/ip layer:
    init_ip_arp_udp_tcp(mymac,myip);

    uart_init(UART_BAUD_SELECT(BAUDRATE,F_CPU));
    sei(); //needed for uart


    while(1){
            // get the next new packet:
            plen = enc28j60PacketReceive(BUFFER_SIZE, buf);

            /*plen will ne unequal to zero if there is a valid 
                * packet (without crc error) */
            if(plen==0){
                    continue;
            }

            // arp is broadcast if unknown but a host may also
            // verify the mac address by sending it to 
            // a unicast address.
            if(eth_type_is_arp_and_my_ip(buf,plen)){
                    make_arp_answer_from_request(buf);
                    continue;
            }

            // check if ip packets are for us:
            if(eth_type_is_ip_and_my_ip(buf,plen)==0){
                    continue;
            }

            if(buf[IP_PROTO_P]==IP_PROTO_ICMP_V && buf[ICMP_TYPE_P]==ICMP_TYPE_ECHOREQUEST_V){
                    // a ping packet, let's send pong
                    make_echo_reply_from_request(buf,plen);
                    continue;
            }

            // udp start, we listen on udp port 1200=0x4B0
            if (buf[IP_PROTO_P]==IP_PROTO_UDP_V && buf[UDP_DST_PORT_H_P]==4 && buf[UDP_DST_PORT_L_P]==0xb0)
            {
                    payloadlen=buf[UDP_LEN_L_P]-UDP_HEADER_LEN;

                    switch (buf[UDP_DATA_P])
                    {
                    case '\1' : //command code = 0x01 -> start of new picture
                        ready = FALSE;
                        packetCount=0;
                        uart_putc('\1');
                        uart_putc(buf[UDP_DATA_P+1]); //tell it to "binem carrier board"
                        break;

                    case '\2' :  //command code = 0x02 -> new line of the picture
                        if(buf[UDP_DATA_P+1] == packetCount)
                        {
                            packetCount++;
                            uart_putc('\2');
                            for(i=2; i<ROW_LEN+2; i++) //ROW_LEN+1 because we send the first command byte before the row's data
                            { uart_putc(buf[UDP_DATA_P+i]); } //tell it to the board & send the data
                            make_udp_reply_from_request(buf, "\6y", 2, 1200);
                        }
                        else
                        { make_udp_reply_from_request(buf, "\6b", 2, 1200); }
                        break;

                    case '\3' : //command code = 0x03 -> end of picture
                        uart_putc('\3'); //tell it to the board (screen will be refreshed to show the picture)
                        ready = TRUE;
                        break;

                    case '\4' : //command code = 0x01 -> start of new picture
                        uart_putc('\4');
                        uart_putc(buf[UDP_DATA_P+1]); //tell it to "binem carrier board"
                        break;

                    case '\5' : //command code = 0x05 -> end of picture
                        uart_putc('\5'); //tell it to the board (screen will be refreshed to show the picture)
                        break;

                    case '\6' : //are you ready ?
                        if (ready == TRUE)
                        { make_udp_reply_from_request(buf, "\6y", 2, 1200); }
                        else
                        { make_udp_reply_from_request(buf, "\6n", 2, 1200); }
                        break;
                    }
            }
    }

    return (0); //will never returns because of the infinite loop
}
