/*********************************************
 * Author: Clément Notin for "Erasme - Département du Rhône" (France) in July 2008
 * Email : clement@notin.org
 * Website : reseau.erasme.org
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Main computer program : load a picture passed as first argument (format is Windows BMP, black & white (indexed color, 1 bit per pixel),
 * there is NO check nor converting : please do this yourself), extract the core data (without headers), and send it via UDP over network
 * to our UDP server's IP (again, no check if it's reachable and alive : your job). UDP has no control mecanism, if a packet is lost, the whole
 * transfer is over and must be restarted.
 *
 *********************************************/

#include <stdio.h>
#include <stdlib.h> //exit()
#include <string.h>
#include <unistd.h> //usleep(), close()
#include "udpcom.h"

#define UDP_SERVER_IP "192.168.0.144" //IP of our UDP server board
#define UDP_PORT_NUM 1200 //port on which our server listens

#define IMAGE_SIZE 19200 //only data, without headers (19200 : HVGA format)
#define IMAGE_FILE_SIZE 19262 //full file size with headers (19262 : HVGA format)
#define BLOCK_LEN 240 //picture is cut in chuncks (called blocks) to be sent over udp, this is the largest usable length

#define TRUE 1
#define FALSE 0

//structure of a BMP image format's header
typedef struct
{
  unsigned short bfType;
  unsigned int   bfSize;
  unsigned short bfReserved1;
  unsigned short bfReserved2;
  unsigned int   bfOffBits;
} bmpFHEAD;

//ask avrserver if ready or not : is busy if another computer is uploading a picture
int serverIsReady(void)
{
    char buf[2]="";
    udpSend("\6?", 2);
    udpReceive(buf);
    if(buf[1]=='y' || buf[1]== '\0')
    { return TRUE; }
    else
    {
        printf("Binem is busy. Maybe another application is chatting with him. Please wait a minute\n");
        return FALSE;
    }
}

//ask avrserver if he has well received the block
int blockIsOk(void)
{
    char buf[2]="";
    udpReceive(buf);
    if(buf[1]=='y' || buf[1]== '\0')
    { return TRUE; }
    else
    { return FALSE; }
}

int main(int argc, char **argv)
{
    printf("Welcome\n--------------\n");
    char picBuf[IMAGE_FILE_SIZE]=""; //will contain the whole image file
    char *startOfBlockPtr=&(picBuf[0]); //pointer to the start of the block in our memory
    //used to create an UDP packet for command 2 (newblock), will contain cmd code, block number & block data
    char bufBeforeSending[BLOCK_LEN+2]="";
    unsigned int block; //used to loop over the blocks
    bmpFHEAD header; //picture's header
    extern int sockfd_s; //socket, defined in udpcom.c
    extern int sockfd_r; //socket, defined in udpcom.c
    char twoBytesCommandsBuf[2]=""; //buffer used to create 2 bytes commands without sprintf

    switch (argc)
    {
        case 2 : //CLEAR
            if (argv[1][0] != 'c')
            { printf("If only 1 parameter is given, the only possible action is to clear the screen, the code is 'c'.\n\
example : \"./binem c\"\n"); exit(1); }

            printf("The screen will be cleared.\n");
            udpInit(UDP_SERVER_IP, UDP_PORT_NUM); //init UDP socket
            if(serverIsReady() == TRUE)
            {
                udpSend("\5a", 2); //send command 5 -> clear screen ('a' is foo because we need at least 2 bytes
                                   //of payload for the udp)
            }
            close(sockfd_s);
            close(sockfd_r);
        break;


        case 3 : //DISPLAY
            if (argv[1][0] != 'd')
            { printf("If 2 parameters are given, the only possible action is to display a picture, the code is 'd'.\n\
example : \"./binem d 1\" to display picture #1\n"); exit(1); }

            printf("The picture #%d will be displayed.\n", atoi(argv[2]));
            udpInit(UDP_SERVER_IP, UDP_PORT_NUM); //init UDP socket

            if(serverIsReady() == TRUE)
            {
                twoBytesCommandsBuf[0]='\4';
                //memory page number from pic number, page start from 0 and pic num from 1
                //1 picture = 3 pages, so we have this small calculation
                twoBytesCommandsBuf[1]=atoi(argv[2])*3-3;
                udpSend(twoBytesCommandsBuf, 2);
            }
            close(sockfd_s);
            close(sockfd_r);
        break;


        case 4 : //SEND
            if (argv[1][0] != 's')
            { printf("If 3 parameters are given, the only possible action is to send a picture, the code is 's'.\n\
example : \"./binem s 6 /home/clem/Desktop/NY.bmp\" : send picture NY.bmp to slot #6\n"); exit(1); }

            printf("The picture %s will be sent to slot #%d.\n", argv[3], atoi(argv[2]));

            FILE *pic = fopen(argv[3], "rb");

            if(pic == NULL)
            { printf("Unable to load picture '%s' !\n", argv[3]); exit(1); }

            fread(&header.bfType, sizeof(header.bfType), 1, pic);
            fread(&header.bfSize, sizeof(header.bfSize), 1, pic);
            fread(&header.bfReserved1, sizeof(header.bfReserved1), 1, pic);
            fread(&header.bfReserved2, sizeof(header.bfReserved2), 1, pic);
            fread(&header.bfOffBits, sizeof(header.bfOffBits), 1, pic);
            fseek(pic, header.bfOffBits, SEEK_SET); //we set the file pointer to the datas' start skipping useless headers

            if(fread(picBuf, 1, IMAGE_SIZE, pic) != IMAGE_SIZE)  //read the whole picture
            { printf("Picture's size is incorrect. Check your picture or IMAGE_SIZE (in the code) !\n"); exit(1); }

            fclose(pic);

            udpInit(UDP_SERVER_IP, UDP_PORT_NUM); //init UDP socket

            if(serverIsReady() == TRUE)
            {
                twoBytesCommandsBuf[0]='\1';
                //memory page number from pic number, page start from 0 and pic num from 1
                //1 picture = 3 pages, so we have this small calculation                
                twoBytesCommandsBuf[1]=atoi(argv[2])*3-3;
                udpSend(twoBytesCommandsBuf, 2);
                sleep(0.1); //dummy tempo

                for(block=0; block<80; block++) //loop on each block
                {
                    //for each block, we send cmd code=2, the block number (used by avrserver for checking)
                    // followed by the block data
                    bufBeforeSending[0]='\2'; //set the command = 0x02 -> new block
                    bufBeforeSending[1]=block; //block number
                    memcpy(&(bufBeforeSending[2]), startOfBlockPtr, BLOCK_LEN); //followed by the block data
                    udpSend(bufBeforeSending, BLOCK_LEN+1); //send the packet created
                    while(blockIsOk() == FALSE) //test if the block have been lost or not
                    {
                        udpSend(bufBeforeSending, BLOCK_LEN+1); //send the packet created
                    }
                    startOfBlockPtr+=BLOCK_LEN; //increment the pointer to the start of the next block
                    printf("%d\n", block);
                }

                sleep(0.1);
                udpSend("\3a", 2); //send command 3 -> end of picture

                printf("Picture sent, exiting...\n");
            }

            close(sockfd_s);
            close(sockfd_r);

        break;


        default:
            printf("Please give me more parameters !\n\
Usage :\n\
-> ./binem c : clear the screen\n\
-> ./binem d 2 : display picture #2\n\
-> ./binem s 6 /home/clem/Desktop/NY.bmp : send picture NY.bmp to slot #6\n"); exit(1);
        break;
    }
    return 0;
}
