/*********************************************
 * Author: Clément Notin for "Erasme - Département du Rhône" (France) in July 2008
 * Email : clement@notin.org
 * Website : reseau.erasme.org
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Main program : initiate serial connection (with Tuxgraphic webserver board) and connection with a binem epaper screen by using BVGP protocol
 * This prog receives commands (3 types) via serial and sends corresponding commands to the epaper, it has some (limited) Hardware Abstraction Layer
 * Be careful : protocol is not fully implemented and there is NO error checking of serial input data nor epaper returning errors
 * Use it at your own risk : you must check what you send to this prog and to the epaper because they offer no protection !
 *
 * Chip type  : ATMEGA162
 *********************************************/

#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h> //sei()
#include <string.h>
#include <util/delay_basic.h>


#include "bits.h" //macros useful for bitwise manipulation
#include "uart.h" //avr usb lib
#include "bvgp.h" //definitions for the Binem Video and Graphic Protocol (v1.0)

#define BAUDRATE 38400 //higher baudrates have been tested : udp server board's clock is not precise enough for a reliable transfer

//settings for this screen (HVGA format 480*320)
#define BLOCK_LEN 240 //picture is cut in chuncks (called blocks) to be sent over udp, this is the largest usable length
#define WIDTH 320 //in pixels
#define SIZE_MSB 0x4B //size of the screen in 2 bytes=480/8 * 320=19200 (=0x4B00) (1 byte=8 pixels !)
#define SIZE_LSB 0

//macro used to pause the program until the screen is ready (= BUSY pin low)
#define wait_while_busy()   while(bit_read(PINB,0) != 0) { }

char buffer[BLOCK_LEN]; //will contain the row received from usb and to be sent to binem
char debugMsg[50]; //buffer to contain the 
uint16_t currentPosition; //used for USB receiving from computer (look usbFunctionSetup() and usbFunctionWrite() )
uint16_t bytesRemaining; //used for USB receiving from computer (look usbFunctionSetup() and usbFunctionWrite() )


//wait "ms" milliseconds
void wait_ms(uint16_t ms)
{
    uint16_t i;
    for(i=0; i<ms; i++)
    {
        _delay_loop_2(3000); //wait 1 ms @ 12Mhz
    }
}


//wait "micros" microseconds (not very accurate)
void wait_micros(uint16_t micros)
{
    uint16_t i;
    for(i=0; i<micros; i++)
    {
        _delay_loop_1(2); //wait about 1 microsec (a little bit more) @ 12Mhz
    }
}


//on this board, portC's MSB is binem's LSB, so 0bGHIJKLMN is reversed as 0bNMLKJIHG
unsigned char reversePinC(unsigned char PinCValue)
{
    unsigned char i,out;
    out=0;
    for(i=0; i<=7; i++)
    {
        if(bit_read(PinCValue, i) != 0)
        { bit_set(out, 7-i); }
    }
    return out;
}


//send a 2 bytes command to binem, be careful : "command" is LSB (sent first), refer to BVGP protocol doc
// speed is critical => optimize ! (maxi timings must be respected)
void sendToScreen(uint8_t command, uint8_t param)
{
    uint8_t PORTCtmp=reversePinC(param);
    wait_while_busy();
    DDRA=255; //PORTA is set to OUTPUT
    DDRC=255; //PORTC is set to OUTPUT
    _delay_loop_1(1); //very short tempo (less than 1 microsec)
    bit_clear(PORTB, 4); //CE
    PORTA=command;
    wait_micros(1);
    PORTC=PORTCtmp;
    wait_micros(1);
    bit_clear(PORTB, 3); //WR
    wait_micros(1);
    bit_set(PORTB, 3); //WR
    _delay_loop_1(1);
    bit_set(PORTB, 4); //CE
}


//receive 2 bytes from binem
//speed is less critical
uint16_t receiveFromScreen(void)
{
    wait_while_busy();
    DDRA=0; //PORTA set to input
    DDRC=0; //PORTC set to input
    _delay_loop_1(1); //very short tempo (less than 1 microsec)
    bit_clear(PORTB, 4); //CE
    _delay_loop_1(1);
    bit_clear(PORTB, 2); //RD
    wait_micros(1);
    uint8_t portAValue=PINA;
    wait_micros(1);
    uint8_t portCValue=PINC;
    bit_set(PORTB, 2); //RD
    _delay_loop_1(1);
    bit_set(PORTB, 4); //CE
    return (reversePinC(portCValue)<<8 | portAValue); //make a 16 bits value with the 2 bytes
}


//small function to retrieve a row content sent over serial
//get one byte at a time and place it into a buffer
void receiveFromSerial(void)
{
    uint8_t i;
    unsigned int value;
    for(i=0; i<BLOCK_LEN; i++)
    {
        do
        {
            value=uart_getc();
        } while (value == UART_NO_DATA); //wait for data
        buffer[i]=value;
    }
}


int main(void)
{
    unsigned int cmd; //1 byte buffer to receive the command code over serial from UDP server

    PORTD = 0xff; //output (security)
    DDRD = 0xF2; //inputs/outputs (cf schematic)
    PORTD = 0xD0; //no pullups (as adviced by AVR USB) on portd except for the fifth bit which must be output to 1 (cf schematic)

    DDRA = 255; //output to screen (less signifiant byte)
    DDRB = 0x3C; //screen's control pins and SPI pins
    DDRC = 255; //output to screen (more signifiant byte)

    PORTA = 0;
    PORTB = 255; //activate internal pullup for SPI inputs, and disable all control input (CE, RD, WR, inactive high)
    PORTC = 0;

    //binem's boot sequence
    bit_set(PORTD, 5);
    wait_ms(120); //100ms mini
    bit_clear(PORTD, 5);

    uart_init(UART_BAUD_SELECT(BAUDRATE,F_CPU));  //initiate serial connection

    wait_while_busy(); //waiting during boot

    sei(); //enable interrupts for serial

    sendToScreen(SET_CONFIG_REG, 0);  //these two lines configure the screen (at each startup, no ROM...)
    wait_ms(1);
    sendToScreen(0b100000, 0); //enable dword alignement and disable "row by row" mode, other bits to 0 (default)


    sendToScreen(SET_CURSOR_POS, 0); //page 0
    sendToScreen(0, 0); //column 0
    sendToScreen(0, 0); //row 0
    wait_ms(100);


    for(;;)
    {
        do
        {
            cmd=uart_getc();
        } while (cmd == UART_NO_DATA); //wait until there is data in serial buffer

        switch (cmd)
        {
        case '\1' : //command code = 0x01 -> start of new picture
            do
            {
                cmd=uart_getc();
            } while (cmd == UART_NO_DATA); //wait until there is data in serial buffer

            sendToScreen(SET_CURSOR_POS, cmd); //page cmd
            sendToScreen(0, 0); //column 0
            sendToScreen(0, 0); //row 0

            sendToScreen(SEND_IMAGE_BLOCK, 0); //command used to start copying picture inside screen's memory
            sendToScreen(SIZE_LSB, SIZE_MSB);

        break;

        case '\2' : //command code = 0x02 -> new block of the picture
            receiveFromSerial(); //get the row's whole data

            for(uint8_t a=0; a<BLOCK_LEN; a+=2) //navigate into the block and send 2 bytes of data each time
            {
                sendToScreen(buffer[a], buffer[a+1]); //send the data
            }

        break;


        case '\3' : //command code = 0x03 -> end of picture
            receiveFromScreen(); //the SEND_IMAGE_BLOCK return a status (but we don't care about it, hope there's no error...)
        break;

        case '\4' :
            do
            {
                cmd=uart_getc();
            } while (cmd == UART_NO_DATA); //wait until there is data in serial buffer

            sendToScreen(DISPLAY_REFRESH, 0b10101); //display picture with bits : clear full, refresh full, page number
            sendToScreen(cmd, 0); //refresh with page number cmd
            receiveFromScreen(); //get status returned by DISPLAY_REFRESH
        break;


        case '\5' :
            sendToScreen(DISPLAY_REFRESH, 1); //display picture with bits : clear full, refresh full, page number
            receiveFromScreen(); //get status returned by DISPLAY_REFRESH
        break;
        }

    }
return 0; //never returns...
}
